@@ -1308,6 +1308,9 @@ extern void json_expand_client_security_groups(json_t *parent, Client *client);
 extern void json_expand_channel(json_t *j, const char *key, Channel *channel, int detail);
 extern void json_expand_tkl(json_t *j, const char *key, TKL *tkl, int detail);
 extern void json_expand_textanalysis(json_t *root, const char *key, TextAnalysis *ta, int detail);
+extern void json_expand_mask_list(json_t *parent, const char *key, ConfigItem_mask *mask);
+extern void json_expand_name_list(json_t *parent, const char *key, NameList *list);
+extern void json_expand_nvplist(json_t *parent, const char *key, NameValuePrioList *list);
 /* end of json.c */
 /* securitygroup.c start */
 extern MODVAR SecurityGroup *securitygroups;
@@ -1490,6 +1493,8 @@ extern long get_connected_time(Client *client);
 extern time_t get_creationtime(Client *client);
 extern const char *StripControlCodes(const char *text);
 extern const char *StripControlCodesEx(const char *text, char *output, size_t outputlen, int strip_flags);
+extern int valid_text(const char *str);
+extern int valid_text_nospaces(const char *str);
 extern MODVAR Module *Modules;
 extern const char *command_issued_by_rpc(MessageTag *mtags);
 extern MODVAR int quick_close;
