@@ -22,6 +22,8 @@
 
 #include "unrealircd.h"
 
+extern struct SetCheck settings;
+
 ISupport *ISupports = NULL; /* List of ISUPPORT (005) tokens */
 ISupport *ISupports_old = NULL; /* see isupport_snapshot() and isupport_check_for_changes() */
 #define MAXISUPPORTLINES 10
@@ -119,6 +121,11 @@ void isupport_init(void)
 	else
 		ISupportDelByName("UHNAMES");
 	ISupportSet(NULL, "DEAF", "d");
+	if (settings.has_network_icon)
+		ISupportSet(NULL, "draft/ICON", iConf.network_icon);
+	else
+		ISupportDelByName("draft/ICON");
+	
 	set_isupport_extban(); /* EXTBAN=xyz */
 	set_isupport_targmax(); /* TARGMAX=... */
 }
@@ -178,10 +185,16 @@ ISupport *ISupportAdd(Module *module, const char *token, const char *value)
 	/* draft-brocklesby-irc-isupport:
 	 * token = a-zA-Z0-9 and 20 or less characters
 	 * value = ASCII 0x21 - 0x7E
+	 *
+	 * Nov 2025:
+	 * Allow also including '/' as per "modern irc"
+	 * https://modern.ircdocs.horse/#rplisupport-005
+	 * https://github.com/ircdocs/modern-irc/issues/250
+	 * - Valware
 	 */
 	for (c = token; c && *c; c++)
 	{
-		if (!isalnum(*c))
+		if (!isalnum(*c) && *c != '/')
 		{
 			if (module)
 				module->errorcode = MODERR_INVALID;
