@@ -90,13 +90,13 @@ void set_user_modes_dont_spread(Client *client, const char *umode)
 {
 	const char *args[4];
 
-	args[0] = client->name;
-	args[1] = client->name;
+	args[0] = NULL;
+	args[1] = client->id;
 	args[2] = umode;
 	args[3] = NULL;
 
 	dontspread = 1;
-	do_cmd(client, NULL, "MODE", 3, args);
+	cmd_umode(NULL, client, NULL, 3, args);
 	dontspread = 0;
 }
 
@@ -834,8 +834,7 @@ void welcome_user(Client *client, TKL *viruschan_tkl)
 	sendnumeric(client, RPL_MYINFO, me.name, version, umodestring, cmodestring);
 
 	RunHook(HOOKTYPE_WELCOME, client, 4);
-	for (i = 0; ISupportStrings[i]; i++)
-		sendnumeric(client, RPL_ISUPPORT, ISupportStrings[i]);
+	send_isupport(client);
 
 	RunHook(HOOKTYPE_WELCOME, client, 5);
 
